#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .. import Node
from .helpers import (
    is_rule_root,
    get_node_label,
    add_to_tree,
    bump_indent
)
from kivish.utils import LineInfo, error
import copy


class TreeBuilder:
    '''
    Builds a node tree for the given code.
    The code must be a list of tuples, featuring num_lines and strings.
    '''

    def __init__(self, root_tree, rules):
        self._root_tree = root_tree
        self._rules = rules

        self._built_rules = {}
        self._previous_walk = {
            'indent': None,
            'inline_attrs': None,
            'value': None
        }
        self._value_stack = []
        self._attr_stack = []
        self._num_line = None

    def go(self):
        '''Returns root node of the code.'''
        root, x = self._traverse_node_tree(self._root_tree)
        return root

    def _traverse_node_tree(self, tree, previous_walk=None):
        self._previous_walk = previous_walk

        root = self._get_tree_root(tree)

        if previous_walk:
            bump_indent(root, previous_walk['indent'])

        return root, self._built_rules

    def _get_tree_root(self, tree):
        created_nodes = []

        for num_line, line in tree:
            self._num_line = num_line

            if not line:
                continue

            node = self._get_node(line, num_line)
            add_to_tree(node, created_nodes)

        return created_nodes[0]

    def _get_node(self, line, num_line):
        if self._node_label_among_rules(line):
            return self._get_or_create_rule(line)
        else:
            return self._create_node(line, num_line)

    def _node_label_among_rules(self, line):
        return get_node_label(line) in self._rules.keys()

    def _get_or_create_rule(self, line):
        info = LineInfo(line)
        current_walk = {
            'inline_attrs': info.inline_attrs,
            'indent': info.indent,
            'value': info.value
        }

        node_label = get_node_label(line)

        if self._rule_already_built(node_label):
            indent = current_walk['indent']
            value = current_walk['value']
            return self._use_existing_rule(node_label, value, indent)
        else:
            return self._create_rule(node_label, current_walk)

    def _use_existing_rule(self, label, value, indent):
        rule = copy.deepcopy(self._built_rules[label])
        rule.value = value
        rule.indent = indent
        return rule

    def _raw_root_rule_node(self, line):
        raw_node = get_node_label(line)

        self._validate_rule_declaration(line)

        if self._attr_stack:
            attrs = attrs_to_string(self._attr_stack)
            raw_node += ' ' + attrs
            self._attr_stack = []

        if self._value_stack:
            raw_node += ': ' + self._value_stack[0]
            self._value_stack = []

        return raw_node

    def _validate_rule_declaration(self, line):
        if len(line.split()) > 1:
            error.report('bad node declaration', self._num_line)

    def _create_node(self, line, num_line):
        if is_rule_root(line):
            raw_node = self._raw_root_rule_node(line)
            return Node(raw_node, num_line)
        else:
            return Node(line, num_line)

    def _rule_already_built(self, line):
        return line in self._built_rules.keys()

    def _create_rule(self, node_label, current_walk):

        if current_walk['value']:
            self._value_stack.append(current_walk['value'])
        if current_walk['inline_attrs']:
            self._attr_stack.append(current_walk['inline_attrs'])

        node, new_built_rules = self._traverse_node_tree(
            self._rules[node_label], current_walk)

        self._built_rules[node_label] = node
        for label, root in new_built_rules.items():
            self._built_rules[label] = root

        return node


def attrs_to_string(attrs_list):

    def get_attrs(attr_set):
        string = ''

        if attr_set['id']:
            string += '#' + attr_set['id']

        class_string = ''
        for class_ in attr_set['classes']:
            class_string += '.' + class_ + ' '
        string += class_string

        return string.rstrip()

    string = ''
    for attr_set in attrs_list:
        string += get_attrs(attr_set)
    return string



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .line_info import LineInfo


def test_indent():
    assert LineInfo('node: ').indent == 0
    assert LineInfo('    node: ').indent == 4


def test_value():
    assert LineInfo('node: val').value == 'val'
    line = 'node: my hobbies: ski, etc.'
    assert LineInfo(line).value == 'my hobbies: ski, etc.'
    line = 'node: my hobbies: ski, etc.'
    assert LineInfo(line).value == 'my hobbies: ski, etc.'
    assert LineInfo('node: \|hi|').value == '|hi|'
    lines = '''node: |
    a
    b'''
    assert LineInfo(lines).value == 'a\nb'


def test_label():
    assert LineInfo('node').label == 'node'
    assert LineInfo('    node:').label == 'node'
    assert LineInfo('    node .x:').label == 'node'


def test_inline_attrs():
    line = 'node #id-x .class-a .class-b:'
    expected = {
        'classes': ['class-a', 'class-b'],
        'id': 'id-x'
    }
    result = LineInfo(line).inline_attrs
    assert expected == result



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INDENT = config['INPUT_INDENT']


class ValueReader:
    '''Reads value of a given line (raw node). Is aware of multiline syntax.'''

    def __init__(self, line):
        self._line = line
        self._line_indent = get_line_indent(line)

    def read(self) -> str:
        '''Return value of the given line.'''
        line, line_indent = self._line, self._line_indent

        value = extract_value_part(line)

        if not value:
            return None

        if is_multiline_value(value):
            return merge_multiline_value(line_indent, value)
        elif has_multiline_escape(value):
            return remove_multiline_escape(value)
        else:
            return value


def extract_value_part(line):
    parts = line.split(':', 1)
    if len(parts) > 1:
        return parts[1].strip()
    else:
        return None


def is_multiline_value(value):
    first_chars = value.lstrip()[:2]
    return first_chars[0] == '|' and first_chars != r'\|'


def merge_multiline_value(node_indent, value):
    inline_value = ''

    parts = value.split('\n')[1:]

    for line in parts:

        line_indent = get_line_indent(line)
        if line_indent == node_indent + INDENT:
            inline_value += line.strip() + '\n'
        else:
            break

    return inline_value.strip()


def get_line_indent(line):
    line = line.rstrip()
    return len(line) - len(line.lstrip())


def has_multiline_escape(value):
    return value.lstrip()[0:2] == '\|'


def remove_multiline_escape(value):
    return value.lstrip()[1:]



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import PrimaryNodesLocator


def test_locate_single_root_line():
    code = PrimaryNodesLocator('root').locate()

    num_line, line = next(code['root'])
    assert num_line == 1
    assert line == 'root'


def test_multiline():
    code = '''
root: |
    hello
    world
'''
    code = PrimaryNodesLocator(code).locate()

    for num_line, line in code['root']:
        if num_line == 2:
            assert line == 'root: hello\nworld'
        elif num_line < 2 or num_line > 2:
            error(num_line, line)


def error(num_line, line):
    raise AssertionError('{}:{}'.format(num_line, line))


def test_locate_multiline_root_and_rules():
    code = '''
root:
    node1
    node2
<rule1@x>:
    node3
<rule2@x>:
    node4:
        node5
'''
    code = PrimaryNodesLocator(code).locate()
    check_root(code)
    check_first_rule(code)
    check_second_rule(code)


def check_root(code):
    for num_line, line in code['root']:
        if num_line == 2:
            assert line == 'root:'
        elif num_line == 3:
            assert line == '    node1'
        elif num_line == 4:
            assert line == '    node2'
        elif num_line < 2 or num_line > 4:
            error(num_line, line)


def check_first_rule(code):
    for num_line, line in code['rules']['rule1']:
        if num_line == 5:
            assert line == '<rule1@x>:'
        elif num_line == 6:
            assert line == '    node3'
        elif num_line < 5 or num_line > 6:
            error(num_line, line)


def check_second_rule(code):
    for num_line, line in code['rules']['rule2']:
        if num_line == 7:
            assert line == '<rule2@x>:'
        elif num_line == 8:
            assert line == '    node4:'
        elif num_line == 9:
            assert line == '        node5'
        elif num_line < 7 or num_line > 9:
            error(num_line, line)



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INDENT = config['INDENT']

from kivish.utils import error


class Analyzer:

    def __init__(self, code):
        self.code = string_to_list(code)

    def check(self):
        '''
        Check for bad indentation and number of roots.

        Raises SyntaxError.
        '''
        roots = 0

        for i, line in enumerate(self.code, 1):

            if not line.strip():
                continue

            if check_bad_indent(line) is True:
                error.report('bad indent', i)

            roots += check_root_node(line)
            if roots > 1:
                error.report('second root', i)

        if roots == 0:
            error.report('no root node')


def string_to_list(string):
    if isinstance(string, list):
        return string

    result = []
    for line in string.split('\n'):
        result.append(line)
    return result


def check_root_node(line):
    if is_significant_line(line) and not is_rule_declaration(line):
        return 1
    return 0


def is_significant_line(line):
    return line[0] != ' ' and line[0] != '#'


def is_rule_declaration(line):
    return line.lstrip()[0] == '<'


def check_bad_indent(line):
    if is_indentation_neutral_line(line):
        return False

    line_indent = get_line_indent(line)

    if line_indent % INDENT != 0 and not is_rule_declaration(line):
        return True

    if is_rule_declaration(line) and line_indent != 0:
        return True


def is_indentation_neutral_line(line):
    return line[0] != ' ' or line.lstrip()[0] == '#'


def get_line_indent(line):
    line = line.rstrip()
    return len(line) - len(line.strip())



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import TreeBuilder
from ..primary_nodes_locator import PrimaryNodesLocator


def test_node_without_value():
    code = 'root:'
    node = get_root_node(code)
    assert node.label == 'root'
    assert node.value is None
    assert node.children == []


def get_root_node(code):
    data = PrimaryNodesLocator(code).locate()
    root, rules = data['root'], data['rules']
    return TreeBuilder(root, rules).go()


def test_node_with_text_value():
    code = 'root: hello world'
    node = get_root_node(code)
    assert node.value == 'hello world'


def test_many_nodes():
    code = '''
root:
    node1: hello
    node2: world
        node3: !
    node4:
'''
    node = get_root_node(code)
    assert node.children[0].label == 'node1'
    assert node.children[0].value == 'hello'
    assert node.children[1].label == 'node2'
    assert node.children[1].value == 'world'
    assert node.children[1].children[0].label == 'node3'
    assert node.children[1].children[0].value == '!'
    assert node.children[2].label == 'node4'


def test_multiline_node():
    code = '''
root: |
    some
    text
    '''
    node = get_root_node(code)
    assert node.value == 'some\ntext'


def test_multiline_nested_gaps():
    code = '''
root:
    node1: |
        a
        b
    node2: |
        c
        d
    '''
    root = get_root_node(code)
    node1 = root.children[0]
    node2 = root.children[1]
    assert node1.num_line == 3
    assert node1.value == 'a\nb'
    assert node2.num_line == 6
    assert node2.value == 'c\nd'


def test_inline_attribute():
    code = '''
root .some-class
'''
    node = get_root_node(code)
    assert node.children[0].label == '*class'
    assert node.children[0].value == 'some-class'


def test_use_rule():
    code = '''
rule
<rule@div>
'''
    node = get_root_node(code)
    assert node.label == 'div'
    assert node.value is None
    assert node.children == []


def test_deny_bad_rule_declaration():
    code = '''rule
<rule@div>: text'''
    try:
        get_root_node(code)
        raise Exception('allowed bad rule declaration')
    except SyntaxError:
        pass

    code = '''rule
<rule@div .x>'''
    try:
        get_root_node(code)
        raise Exception('allowed bad rule declaration')
    except SyntaxError:
        pass


def test_rule_with_attributes():
    code = '''
rule .some-class
    *width: 1
<rule@div>
    '''
    node = get_root_node(code)
    assert node.children[0].label == '*class'
    assert node.children[0].value == 'some-class'
    assert node.children[1].label == '*width'
    assert node.children[1].value == '1'


def test_use_nested_rules():
    code = '''
d1
    a
    d2
        d3

<d1@div>
<d2@span>
<d3@h1>
'''
    node = get_root_node(code)
    assert node.label == 'div'
    a = node.children[0]
    assert a.label == 'a'
    d2 = node.children[1]
    assert d2.label == 'span'
    d3 = d2.children[0]
    assert d3.label == 'h1'


def test_nested_rule_with_attribute():
    code = '''
node
    nested

<nested@div>:
    *test: passed
'''
    node = get_root_node(code)
    assert node.children[0].children[0].value == 'passed'


def test_rule_inheritance():
    code = '''
rule1 .a: hello world
    *style: color: blue

<rule1@rule2>
    *class: b
    test1: val_1

<rule2@div>
    *class: c
    *style: color: red
    test2: val_2
    '''
    node = get_root_node(code)
    assert node.label == 'div'
    assert node.value == 'hello world'

    vals = ('val_1', 'val_2', 'a c b', 'color: red; color: blue')
    for val in vals:
        assert val in [child.value for child in node.children]


def test_rule_used_in_many_places():
    code = '''
root:
    TheRule: hello
    div:
        TheRule: world

<TheRule@span>
'''
    root = get_root_node(code)
    assert root.children[0].value == 'hello'
    assert root.children[1].children[0].value == 'world'



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .analyzer import Analyzer
from ..node import TreeBuilder, PrimaryNodesLocator


class Scanner:
    '''Scans a kvs code to build a tree of nodes.'''

    def __init__(self, code):
        Analyzer(code).check()
        self.code = code

    def scan(self):
        '''Builds a tree of nodes and returns a root node.'''
        data = PrimaryNodesLocator(self.code).locate()
        root_tree, rules = data['root'], data['rules']
        return TreeBuilder(root_tree, rules).go()



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import Node


def test_no_colon():
    node = Node('Root-node-1', 0)
    assert node.label == 'Root-node-1'
    assert node.value is None


def test_colon_in_value():
    node = Node('root: hello:world:', 0)
    assert node.value == 'hello:world:'


def test_multiline_value():
    code = '''root: |
    long multiline
    text
    '''
    node = Node(code, 0)
    assert node.value == 'long multiline\ntext'


def test_multiline_escape():
    node = Node('root: \|hello\|', 0)
    assert node.value == '|hello\|'


def test_quoted_value():
    node = Node('root: "hello world"', 0)
    assert node.value == '"hello world"'
    node = Node("root: 'hello world'", 0)
    assert node.value == "'hello world'"


def test_attribute_node():
    node = Node('*node:', 0)
    assert node.label == '*node'


def test_deny_bad_label():
    try:
        Node('spaced node:', 0)
        Node('$node:', 0)
        Node('node_:', 0)
        raise Exception('bad name accepted')
    except SyntaxError:
        pass


def test_inline_attribute():
    node = Node('node #id-x .class-x .class-y', 0)
    expected_values = ('id-x', 'class-x class-y')

    assert False if len(node.children) == 0 else True
    for child in node.children:
        assert child.value in expected_values



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
import re

name = '[A-Za-z0-9-]+'
star_or_slash = '(?:\*?|\/?)'

pattern_label = re.compile('^{} *({})$'.format(star_or_slash, name))
pattern_class = re.compile('\.({})'.format(name))
pattern_id = re.compile('#({})'.format(name))



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import Generator
from ..node import Node
from ..scanner import Scanner
from pytest import fixture


@fixture
def simple_root():
    root = Node('html', 0)
    head = Node('    head', 1)
    body = Node('    body: hello world', 2)
    root.add_child(head)
    root.add_child(body)
    return root


def test_generate_simple_size(simple_root):
    html = Generator(simple_root).generate()

    expected = [
        '<html>',
        '  <head>',
        '  </head>',
        '  <body>',
        '    hello world',
        '  </body>',
        '</html>'
    ]
    assert_sequence(html, expected)


def assert_sequence(sequence, expected_sequence):
    sequence = sequence.split('\n')
    for i in range(len(expected_sequence)):
        assert sequence[i] == expected_sequence[i]


def test_attributes():
    node = Node('node #1 .x', 0)
    attr = Node('    *style: color: red', 0)
    node.add_child(attr)
    html = Generator(node).generate()

    expected = [
        '<node class="x" id="1" style="color: red">',
        '</node>'
    ]
    assert_sequence(html, expected)


def test_inline_node():
    node = Node('/img', 0)
    src = Node('*src: link', 0)
    node.add_child(src)
    expected = '<img src="link"/>\n'
    html = Generator(node).generate()
    assert html == expected


def test_nested_rules():
    code = '''
html
    body
        rule .a

<rule@x>
    *class: b
    /img:
        *src: link_b

<x@div>
    *class: c
    /img:
        *src: link_c
    '''
    node = Scanner(code).scan()
    html = Generator(node).generate()

    expected = [
        '<html>',
        '  <body>',
        '    <div class="a c b">',
        '      <img src="link_c"/>',
        '      <img src="link_b"/>',
        '    </div>',
        '  </body>',
        '</html>',
    ]
    assert_sequence(html, expected)


def test_deny_attribute_with_children():
    code = '''
node:
    *class:
        x:
'''
    node = Scanner(code).scan()
    try:
        Generator(node).generate()
        raise Exception('accepted bad node')
    except SyntaxError:
        pass


def test_deny_bad_inline_node():
    code = '''
/node:
    x:
'''
    node = Scanner(code).scan()
    try:
        Generator(node).generate()
        raise Exception('accepted bad inline node')
    except SyntaxError:
        pass


def test_flat_attributes():
    code = '''
node:
    *style: |
        a;
        b
'''
    node = Scanner(code).scan()
    html = Generator(node).generate()
    assert 'style="a; b"' in html


def test_attribute_without_value():
    code = '''
node:
    *flex
'''
    node = Scanner(code).scan()
    html = Generator(node).generate()
    assert '<node flex>' in html


def test_attribute_without_value_inline():
    code = '''
/node:
    *flex
'''
    node = Scanner(code).scan()
    html = Generator(node).generate()
    assert '<node flex/>\n' == html


def test_multiline_indentation():
    code = '''
node: |
    hello
    world
'''
    node = Scanner(code).scan()
    html = Generator(node).generate()
    expected = [
        '<node>',
        '  hello',
        '  world',
        '</node>'
    ]
    assert_sequence(html, expected)



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-


def report(msg, num_line=None):
    if num_line:
        raise SyntaxError('<error @ line {}> {}'.format(num_line, msg))
    else:
        raise SyntaxError('<error> {}'.format(msg))



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .regex import pattern_label, pattern_class, pattern_id
from .helper_value_reader import ValueReader
from . import error


class LineInfo:
    '''
    Gives useful information about a given line of code, such as:
    - indent (number of spaces)
    - label (string; validated, might raise SyntaxError)
    - value (string)
    - inline attributes (list of tuples)
    '''

    @property
    def indent(self):
        line = self._line.rstrip()
        return len(line) - len(line.lstrip())

    @property
    def value(self):
        return ValueReader(self._line).read()

    @property
    def label(self):
        if not self._label:
            self._label = self._line.lstrip().split()[0].rstrip(':')
            self._check_label()
            return self._label
        else:
            return self._label

    def _check_label(self):
        '''Raises SyntaxError on failure.'''
        if not pattern_label.match(self.label):
            error.report('bad label (%s)' % self.label)

    @property
    def inline_attrs(self):
        '''
        Returns a dict with classes and id. Example:
        {
            'classes': ['class-a', class-b],
            'id': 'id-x'
        }
        '''
        return get_inline_classes_and_id(self._line)

    def __init__(self, line):
        self._line = line
        self._label = None


def get_inline_classes_and_id(line):
    classes = pattern_class.findall(line)
    id = pattern_id.findall(line)
    id = id[-1] if id else None
    return {'classes': classes, 'id': id}



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INDENT = config['INDENT']


def string_to_list(string):
    '''
    Rewrites given code into a list. If it contains multiline nodes,
    merges them into single logical lines.
    '''
    code = []

    multiline = {'indent': None, 'parts': []}

    for line in string.split('\n'):

        if reading_multiline_in_progress(multiline):
            finished = read_multiline(line, multiline)

            if not finished:
                continue
            else:
                code.append(finished)

        if is_multiline_label(line):
            multiline['indent'] = get_line_indent(line)
            multiline['parts'].append(get_node_label(line))
            continue

        code.append(line)

    return code


#################################


def is_multiline_label(line):
    return line.rstrip().endswith('|')


def get_line_indent(line):
    line = line.rstrip()
    return len(line) - len(line.lstrip())


def get_node_label(code):
    return code.split(':')[0] + ':'


def reading_multiline_in_progress(multiline):
    return multiline['indent'] is not None


def read_multiline(line, multiline):
    '''Returns node if finished reading, else False.'''
    if is_multiline_part(multiline['indent'], line):
        multiline['parts'].append(line)
        return False

    return merge_multiline_parts(multiline)


def is_multiline_part(multiline_indent, line):
    return multiline_indent <= get_line_indent(line) - INDENT


def merge_multiline_parts(multiline):
    indent = multiline['indent']
    merged = multiline['parts'][0] + ' '

    for part in multiline['parts'][1:]:
        merged += extract_indented_text(part, indent)

    multiline['indent'], multiline['parts'] = None, []

    return merged[:-1]


def extract_indented_text(part, indent):
    return part[indent + INDENT:].rstrip() + '\n'



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
INDENT = 2

from kivish.utils import error, regex


class Generator:
    '''Creates an html file from a provided kvs root node.'''

    def __init__(self, root_node):
        self._root_node = root_node

    def generate(self):
        '''Creates an html string and returns it.'''
        html = node_to_string(self._root_node)
        return html


def node_to_string(node):
    if is_inline_node(node):
        return create_inline_node(node)
    else:
        return create_normal_node(node)


def get_spaces(indent):
    if indent == 0:
        return ''
    else:
        return ' ' * int(indent / 2)


def is_inline_node(node):
    return node.label[0] == '/'


def create_inline_node(node):
    spaces = get_spaces(node.indent)
    label = regex.pattern_label.findall(node.label)[0]

    string = spaces + '<' + label

    for child in node.children:
        if is_attribute_node(child):
            string += get_attribute(child)
        else:
            msg = 'inline node can have only attribute children'
            error.report(msg, node.num_line)

    string += '/>\n'
    return string


def is_attribute_node(node):
    if node.label[0] == '*':
        if node.children:
            msg = 'attribute node cannot have children'
            error.report(msg, node.num_line)
        return True


def get_attribute(child):
    if child.value:
        child.value = child.value.replace('\n', ' ')
        return ' ' + child.label[1:] + '="' + child.value + '"'
    else:
        return ' ' + child.label[1:]


def create_normal_node(node):
    spaces = get_spaces(node.indent)
    label = regex.pattern_label.findall(node.label)[0]

    string = spaces + '<' + label + '>\n'

    for child in bump_attribute_nodes(node.children):
        if is_attribute_node(child):
            string = embed_attribute(child, string)
        else:
            string += node_to_string(child)

    string = add_value(string, node, spaces)

    string += spaces + '</' + label + '>\n'
    return string


def bump_attribute_nodes(children):
    que = []

    for child in children:
        if child.label[0] == '*':
            yield child
        else:
            que.append(child)

    for child in que:
        yield child


def embed_attribute(child, string):
    string = string[:-2]
    if child.value:
        child.value = child.value.replace('\n', ' ')
        string += ' ' + child.label[1:] + '="' + child.value + '">\n'
    else:
        string += ' ' + child.label[1:] + '>\n'
    return string


def add_value(string, node, spaces):
    if node.value:

        parts = node.value.split('\n')
        for part in parts:
            string += spaces + ' ' * INDENT + part + '\n'

    return string



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .string_to_list import string_to_list


class PrimaryNodesLocator:

    def __init__(self, code: str):
        self.code = string_to_list(code)

    def locate(self) -> dict:
        '''
        Locate root and rules in the given code.

        Returns a dict with generators of code for each entity.
        Example structure:
        {
            'root': code_generator
            'rules': {
                'some_label': code_generator
                'another_label': code_generator
            }
        }

        Each generator iteration returns a tuple of code and original
        line number. Example structure: (1, 'root:')
        '''
        code = self.code

        root = None
        rules = {}

        for num_line, line in enumerate(code):

            if not line:
                continue
            elif is_root(line):
                root = traverse_node_tree(num_line, code)
            elif is_rule(line):
                label = get_rule_label(line)
                rules[label] = traverse_node_tree(num_line, code)

        return {'root': root, 'rules': rules}


def is_root(line):
    return line[0] not in ' #<'


def traverse_node_tree(start_line, code):
    code = code[start_line:]
    num_line = start_line + 1

    yield num_line, code[0]
    num_line += count_extra_node_height(code[0])

    for line in code[1:]:
        num_line += 1

        if not line:
            continue
        elif not is_root(line) and not is_rule(line):
            yield num_line, line.rstrip()
            num_line += count_extra_node_height(line)
        else:
            break


def count_extra_node_height(code):
    '''Multiline nodes take more than a single line of code.'''
    height = code.count('\n')
    if height > 0:
        height += 1
    return height


def is_rule(line):
    return line[0] == '<'


def get_rule_label(string):
    return string.split('@')[0][1:]



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INDENT = config['INDENT']

from kivish.utils import LineInfo


class Node:
    '''
    A node which represents one or more lines of code, which has:

    - label (string)
    - value (string)
    - num_line (number; not unique for inline attribute nodes)
    - indent (number of spaces)
    - children (child nodes)

    Raises SyntaxError if its label contains forbidden characters.
    '''

    def __init__(self, code, num_line):
        self.label = None
        self.value = None
        self.indent = None
        self.num_line = None
        self.children = []

        self.__set_pools(code, num_line)

    def __set_pools(self, code, num_line):
        code = merge_lines(code)
        info = LineInfo(code)

        self.label = info.label
        self.value = info.value
        self.indent = info.indent
        self.num_line = num_line

        info = LineInfo(code)
        attrs = info.inline_attrs

        if has_inline_attrs(code, attrs):
            indent = info.indent

            for node in get_attr_nodes(attrs, indent, num_line):
                self.add_child(node)

    def add_child(self, node):
        if node.label == '*class':
            if self._already_has_class():
                self._class_node.value += ' ' + node.value
                return

        self.children.append(node)

    def _already_has_class(self):
        return '*class' in [c.label for c in self.children]

    @property
    def _class_node(self):
        for child in self.children:
            if child.label == '*class':
                return child


def merge_lines(code):
    if isinstance(code, str):
        return code
    elif isinstance(code, list):
        return '\n'.join(code)


def has_inline_attrs(code, attrs):
    first_char = code.lstrip()[0]
    if first_char != '*':
        return attrs['classes'] or attrs['id']


def get_attr_nodes(attrs, indent, num_line):
    raw_nodes = create_raw_attribute_nodes(attrs, indent)
    for raw_node in raw_nodes:
        yield Node(raw_node, num_line)


def create_raw_attribute_nodes(attributes, indent):
    classes = attributes['classes']
    id = attributes['id']
    spaces = ' ' * (indent + INDENT)

    list = []

    for class_ in classes:
        list.append('{}*class: {}'.format(spaces, class_))

    if id:
        list.append('{}*id: {}'.format(spaces, id))

    return list



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .scanner import Scanner
from .generator import Generator


def compile(target):
    if not target.endswith('.kvs'):
        return print('Expected a file with .kvs extension.')

    code = read_code(target)

    try:
        nodes = Scanner(code).scan()
    except SyntaxError as e:
        return print(e)

    try:
        html = Generator(nodes).generate()
    except SyntaxError as e:
        return print(e)

    write_html(target, html)


def read_code(target):
    code = ''
    with open(target) as f:
        for line in f:
            code += line
    return code


def write_html(target, string):
    new_target = target.replace('kvs', 'html')
    with open(new_target, 'w') as f:
        f.write(string)



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import Analyzer


def test_no_code():
    should_fail([], 'no code')


def should_fail(code, text):
    try:
        Analyzer(code).check()
    except SyntaxError:
        return
    raise SyntaxError(text)


def test_deny_no_root():
    should_fail('    line', 'no root')
    should_fail('<line@x>', 'no root')


def test_lines():
    code = '''
line
    '''
    Analyzer(code).check()


def test_accept_multiline():
    code = '''
line
    line
    line'''
    Analyzer(code).check()


def test_deny_multiple_root():
    code = '''
root
root'''
    should_fail(code, 'multiple root')


def test_deny_bad_indent():
    code = '''
line
   line'''
    should_fail(code, 'bad indent')


def test_accept_comment():
    code = '''
line
#'''
    Analyzer(code).check()


def test_accept_poorly_indented_comment():
    code = '''
line
 #'''
    Analyzer(code).check()


def test_accept_rule():
    code = '''
root
<rule@x>'''
    Analyzer(code).check()


def test_deny_indented_rule():
    code = '''
root
    <rule@x>'''
    should_fail(code, 'indented rule')



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .string_to_list import string_to_list


def test_single_multiline():
    code = '''
root: |
    selector {
        color: red
    }
'''
    code_list = string_to_list(code)
    assert code_list[1] == 'root: selector {\n    color: red\n}'


def test_many_nodes_and_multiline():
    code = '''
root:
    node1: |
        hello
        world
    node2
'''
    code_list = string_to_list(code)
    assert code_list[0] == ''
    assert code_list[1] == 'root:'
    assert code_list[2] == '    node1: hello\nworld'
    assert code_list[3] == '    node2'
    assert code_list[4] == ''



from setuptools import setup, find_packages

setup(
    name="kivish",
    description='html template language',
    version="0.1.1",
    packages=find_packages(),
    scripts=['kivish/scripts/kivish'],
    author='Jakub Ligęza',
    author_email='jakub.jaroslaw.ligeza@gmail.com',
    license='MIT',
    url='https://gitlab.com/jligeza/kivish',
)



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
__all__ = ['Node', 'TreeBuilder', 'PrimaryNodesLocator']
from .main import Node
from .tree_builder import TreeBuilder
from .primary_nodes_locator import PrimaryNodesLocator



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INDENT = config['INDENT']

from kivish.utils import error


def is_rule_root(line):
    return line[0] == '<'


def get_node_label(line):
    if '@' in line.split(':')[0]:
        return line.split('@')[1].split('>')[0]
    return line.split(':')[0].split()[0].lstrip()


def add_to_tree(node, created_nodes):
    if not created_nodes:
        created_nodes.append(node)
    else:
        add_to_parent(node, created_nodes)


def add_to_parent(new_node, created_nodes):
    '''
    Add the new node as a child of a previously created
    node, which has smaller indentation by one level.
    '''
    for node in reversed(created_nodes):
        if indentation_matches(node, new_node):
            result = is_duplicated_attribute_node(node, new_node)
            if result:
                merge_nodes(result, new_node)
            else:
                node.add_child(new_node)
                created_nodes.append(new_node)
            return

    msg = 'no parent for node "{%s}"' % new_node.label
    error.report(msg, new_node.num_line)


def indentation_matches(node, new_node):
    return node.indent == new_node.indent - INDENT


def is_duplicated_attribute_node(parent, node):
    if node.label[0] == '*':
        for child in parent.children:
            if child.label == node.label:
                return child
    return False


def merge_nodes(parent, new_node):
    if 'class' in new_node.label:
        parent.value += ' ' + new_node.value
    elif 'style' in new_node.label:
        parent.value += '; ' + new_node.value


def bump_indent(node, extra_indent):
    '''
    Used to bump indent of rule nodes to match indent of nodes,
    which use them.
    '''
    node.indent += extra_indent
    for child in node.children:
        bump_indent(child, extra_indent)
